% Code for Example 4.4 in the manuscript.
%
% This example focuses on the running time of the
% structured algorithm w.r.t. system size.
%
% The test matrices are defined by random
% quasiseparable generators of order 3.
%
% This example is meant to show that:
% - the running time grows linearly with n,
% - the structured method presented in the paper
%   is faster than the "naive" sequential method
%   by a factor 2.
%

% Set up structure parameters.

M=20;  % number of shifts
rL=3; % lower qs order
rU=3; % upper qs order
Ntrials=10;

% Preallocate vectors for timings.
time_fast=zeros(1,Ntrials);
time_slow=zeros(1,Ntrials);

% begin loop over Ntrials trials for increasing sizes of A
for N_counter=1:Ntrials
    N=100*N_counter; % size of A
    
    % form random quasiseparable generators
    p=[zeros(1,rL);rand(N-1,rL)]; % N x rL (but first row not used)
    q=rand(rL,N-1);   % rL x (N-1)
    a=rand(rL,rL,N-1); % rL x rL x (N-1)  but first matrix not used
    g=rand(N-1,rU); % (N-1) x rU
    h=[zeros(rU,1),rand(rU,N-1)];   % rU x N but first column not used
    b=rand(rU,rU,N-1); % rU x rU x (N-1)  but first matrix not used
    d=rand(N,1);
    y=rand(N,1); % rhs vectors
    sigma=1i*rand(M,1);  % shifts
    w=zeros(N,1);
    
    % "Fast" version
    
    tic
    % Initialize matrix of solutions.
    x=zeros(N,M);
    
    
    % Implement algorithm.
    
    % Start with the factorization A + sigma I = V * T:
    % initialize generators for V
    pV=zeros(N,rL);
    qV=zeros(rL,N-1);
    aV=zeros(rL,rL,N-1);
    gV=zeros(N-1,rU);
    hV=zeros(rU,N);
    bV=zeros(rU,rU,N-1);
    dV=zeros(N,1);
    
    % 1 %
    
    % Compute sizes:
    m=ones(N,1); % block sizes
    rho=zeros(N,1);
    rhop=zeros(N,1);
    nu=zeros(N,1);
    for k=N:-1:2
        rho(k-1)=min(m(k)+rho(k),rL);
    end
    for k=1:N-1
        rhop(k)=rho(k)+rU;
    end
    nu(1)=m(1)+rho(1);
    for k=2:N
        nu(k)=m(k)+rho(k)-rho(k-1);
    end
    
    gT=zeros(N-1,rhop(1));
    hT=zeros(rhop(1),N);
    bT=zeros(rhop(1),rhop(1),N-1);
    dT=zeros(N,1);
    
    V_N=1;
    Xold=p(N,:);
    pV(N,1)=1;
    hT(1:rhop(N-1),N)=[h(:,N);V_N'*d(N)];
    
    c=y(N);
    gamma_N=[zeros(rU,1);pV(N,1)'];
    gamma=zeros(2*rU,N-1);
    
    for k=N-1:-1:2
        % 1.2.1
        emme=[p(k,:);Xold*a(:,:,k)];
        [V,R]=qr(emme);
        
        Xnew=R(1:rho(k-1),:);
        pV(k,1:rho(k-1))=V(1,1:rho(k-1));
        aV(1:rho(k),1:rho(k-1),k)=V(2:2+rho(k)-1,1:rho(k-1));
        if rho(k)+1>rho(k-1)
            dV(k)=V(1,rho(k-1)+1:rho(k-1)+nu(k));
            qV(:,k)=V(2:2+rho(k)-1,rho(k-1)+1);
        end
        
        %1.2.2
        W=[h(:,k),b(:,:,k),zeros(rU,rho(k));d(k),g(k,:),zeros(1,rho(k));...
            Xold*q(:,k),zeros(rho(k),rU),eye(rho(k))];
        Z=[eye(rU),zeros(rU,rho(k)+1);zeros(rho(k)+1,rU),V']*W; % size (rU+rho(k)+1)x(rU+rho(k)+1)
        hT(1:rhop(k-1),k)=Z(1:rhop(k-1),1);
        bT(1:rhop(k-1),1:rhop(k),k)=Z(1:rhop(k-1),2:rhop(k)+1);
        if (rU+rho(k)+1)>rhop(k-1)
            dT(k)=Z(rhop(k-1)+1,1);
            gT(k,:)=Z(rhop(k-1)+1,2:rhop(k)+1);
        end
        
        gamma(:,k)=[zeros(rU,1);pV(k,:)'];
        auxv=V'*[y(k);c];
        c=auxv(1:rho(k-1));
        w(k)=auxv(end);
        Xold=Xnew;
    end
    % 1.3 (larger sizes here)
    V1=eye(nu(1));
    dT1=[d(1);Xnew*q(:,1)];
    gT1=[g(1,:) zeros(1,rho(1)); zeros(rho(1),rU) eye(rho(1))];
    gamma1=[1;zeros(rho(1),1)];
    w1=V1'*[y(1);c];  %3x1
    
    % 2 %
    
    for i=1:M   % for each shifted system
        
        dR=zeros(N,1);
        gR=zeros(N-1,max(rhop));
        v=zeros(N,1);
        
        % 2.1.1
        [U,R]=qr(dT1+sigma(i)*gamma1);
        dR(1)=R(1,1);
        
        auxv=U'*gT1;
        auxw=U'*w1;
        gR(1,1:rhop(1))=auxv(1,:);
        Y=auxv(2:end,:);
        v(1)=auxw(1);
        alpha=auxw(2:end);
        
        % 2.1.2
        
        for k=2:N-1
            if (rU+rho(k)+1)>rhop(k-1)  % that is, if dT(k) and dV(k) are not empty
                emme=[Y*(hT(:,k)+sigma(i)*gamma(:,k));dT(k)+sigma(i)*dV(k).'];
            else
                emme=Y*(hT(1:rhop(k-1),k)+sigma(i)*gamma(1:rhop(k-1),k));
            end
            [U,R]=qr(emme);
            dR(k)=R(1,1);
            
            if (rU+rho(k)+1)>rhop(k-1) % that is, if gT is not empty
                auxv=U'*[Y*bT(:,:,k); gT(k,:)];
                auxw=U'*[alpha; w(k)];
            else
                auxv=U'*Y*bT(1:rhop(k-1),1:rhop(k),k);
                auxw=U'*alpha;
            end
            
            gR(k,1:rhop(k))=auxv(1,1:rhop(k));
            v(k)=auxw(1);
            Y=auxv(2:1+rho(k),1:rhop(k));
            alpha=auxw(2:end);
        end
        
        %2.1.3
        [U,R]=qr(Y*(hT(1:rhop(N-1),N)+sigma(i)*gamma_N));
        dR(N)=R(1);
        v(N)=U'*alpha;
        
        %2.2 (compute solution x_i=x(:,i))
        
        %2.2.1
        eta=zeros(rhop(1),N-1);
        x(N,i)=dR(N)\v(N);
        eta(1:rhop(N-1),N-1)=(hT(1:rhop(N-1),N)+sigma(i)*gamma_N)*x(N,i);
        %2.2.2
        
        for k=N-1:-1:2
            x(k,i)=dR(k)\(v(k)-gR(k,1:rhop(k))*eta(1:rhop(k),k));
            eta(1:rhop(k-1),k-1)=bT(1:rhop(k-1),1:rhop(k),k)*eta(1:rhop(k),k)+...
                (hT(1:rhop(k-1),k)+sigma(i)*gamma(1:rhop(k-1),k))*x(k,i);
        end
        %2.2.3
        x(1,i)=dR(1)\(v(1)-gR(1,:)*eta(:,1));
    end
    time_fast(N_counter)=toc;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % "Slow" version
    
    tic
    % Initialize matrix of solutions.
    x=zeros(N,M);
    
    for i=1:M   % for each shifted system
        % Implement algorithm.
        
        % Start with the factorization A + sigma I = V * T:
        % initialize generators for V
        pV=zeros(N,rL);
        qV=zeros(rL,N-1);
        aV=zeros(rL,rL,N-1);
        gV=zeros(N-1,rU);
        hV=zeros(rU,N);
        bV=zeros(rU,rU,N-1);
        dV=zeros(N,1);
        
        % 1 %
        
        % Compute sizes:
        m=ones(N,1); % block sizes
        rho=zeros(N,1);
        rhop=zeros(N,1);
        nu=zeros(N,1);
        for k=N:-1:2
            rho(k-1)=min(m(k)+rho(k),rL);
        end
        for k=1:N-1
            rhop(k)=rho(k)+rU;
        end
        nu(1)=m(1)+rho(1);
        for k=2:N
            nu(k)=m(k)+rho(k)-rho(k-1);
        end
        
        gT=zeros(N-1,rhop(1));
        hT=zeros(rhop(1),N);
        bT=zeros(rhop(1),rhop(1),N-1);
        dT=zeros(N,1);
        
        V_N=1;
        Xold=p(N,:);
        pV(N,1)=1;
        hT(1:rhop(N-1),N)=[h(:,N);V_N'*d(N)];
        
        c=y(N);
        gamma_N=[zeros(rU,1);pV(N,1)'];
        gamma=zeros(2*rU,N-1);
        
        for k=N-1:-1:2
            % 1.2.1
            emme=[p(k,:);Xold*a(:,:,k)];
            [V,R]=qr(emme);
            
            Xnew=R(1:rho(k-1),:);
            pV(k,1:rho(k-1))=V(1,1:rho(k-1));
            aV(1:rho(k),1:rho(k-1),k)=V(2:2+rho(k)-1,1:rho(k-1));
            if rho(k)+1>rho(k-1)
                dV(k)=V(1,rho(k-1)+1:rho(k-1)+nu(k));
                qV(:,k)=V(2:2+rho(k)-1,rho(k-1)+1);
            end
            
            %1.2.2
            W=[h(:,k),b(:,:,k),zeros(rU,rho(k));d(k),g(k,:),zeros(1,rho(k));...
                Xold*q(:,k),zeros(rho(k),rU),eye(rho(k))];
            Z=[eye(rU),zeros(rU,rho(k)+1);zeros(rho(k)+1,rU),V']*W; % size (rU+rho(k)+1)x(rU+rho(k)+1)
            hT(1:rhop(k-1),k)=Z(1:rhop(k-1),1);
            bT(1:rhop(k-1),1:rhop(k),k)=Z(1:rhop(k-1),2:rhop(k)+1);
            if (rU+rho(k)+1)>rhop(k-1)
                dT(k)=Z(rhop(k-1)+1,1);
                gT(k,:)=Z(rhop(k-1)+1,2:rhop(k)+1);
            end
            
            gamma(:,k)=[zeros(rU,1);pV(k,:)'];
            auxv=V'*[y(k);c];
            c=auxv(1:rho(k-1));
            w(k)=auxv(end);
            Xold=Xnew;
        end
        % 1.3 (larger sizes here)
        V1=eye(nu(1));
        dT1=[d(1);Xnew*q(:,1)];
        gT1=[g(1,:) zeros(1,rho(1)); zeros(rho(1),rU) eye(rho(1))];
        gamma1=[1;zeros(rho(1),1)];
        w1=V1'*[y(1);c];  %3x1
        
        % 2 %
        
        
        
        dR=zeros(N,1);
        gR=zeros(N-1,max(rhop));
        v=zeros(N,1);
        
        
        
        % 2.1.1
        [U,R]=qr(dT1+sigma(i)*gamma1);
        dR(1)=R(1,1);
        
        auxv=U'*gT1;
        auxw=U'*w1;
        gR(1,1:rhop(1))=auxv(1,:);
        Y=auxv(2:end,:);
        v(1)=auxw(1);
        alpha=auxw(2:end);
        
        % 2.1.2
        
        for k=2:N-1
            if (rU+rho(k)+1)>rhop(k-1)  % that is, if dT(k) and dV(k) are not empty
                emme=[Y*(hT(:,k)+sigma(i)*gamma(:,k));dT(k)+sigma(i)*dV(k).'];
            else
                emme=Y*(hT(1:rhop(k-1),k)+sigma(i)*gamma(1:rhop(k-1),k));
            end
            [U,R]=qr(emme);
            dR(k)=R(1,1);
            
            if (rU+rho(k)+1)>rhop(k-1) % that is, if gT is not empty
                auxv=U'*[Y*bT(:,:,k); gT(k,:)];
                auxw=U'*[alpha; w(k)];
            else
                auxv=U'*Y*bT(1:rhop(k-1),1:rhop(k),k);
                auxw=U'*alpha;
            end
            
            gR(k,1:rhop(k))=auxv(1,1:rhop(k));
            v(k)=auxw(1);
            Y=auxv(2:1+rho(k),1:rhop(k));
            alpha=auxw(2:end);
        end
        
        %2.1.3
        [U,R]=qr(Y*(hT(1:rhop(N-1),N)+sigma(i)*gamma_N));
        dR(N)=R(1);
        v(N)=U'*alpha;
        
        %2.2 (compute solution x_i=x(:,i))
        
        %2.2.1
        eta=zeros(rhop(1),N-1);
        x(N,i)=dR(N)\v(N);
        eta(1:rhop(N-1),N-1)=(hT(1:rhop(N-1),N)+sigma(i)*gamma_N)*x(N,i);
        %2.2.2
        
        for k=N-1:-1:2
            x(k,i)=dR(k)\(v(k)-gR(k,1:rhop(k))*eta(1:rhop(k),k));
            eta(1:rhop(k-1),k-1)=bT(1:rhop(k-1),1:rhop(k),k)*eta(1:rhop(k),k)+...
                (hT(1:rhop(k-1),k)+sigma(i)*gamma(1:rhop(k-1),k))*x(k,i);
        end
        %2.2.3
        x(1,i)=dR(1)\(v(1)-gR(1,:)*eta(:,1));
    end
    time_slow(N_counter)=toc;
end

ratio=time_slow./time_fast;

% Output ratios:
fprintf('%s   %s  %s  %s\n','size','time_fast','time_slow','ratio')
for j=1:Ntrials
    fprintf('%4d %10.2e %10.2e %10.2e\n',100*j,time_fast(j),time_slow(j),ratio(j))
end

% Plot timings for fast method with linear fit:
figure(1)
title('Ex. 4.4')
xlabel('matrix size')
ylabel('time (in seconds)')
xdata=100*(1:Ntrials);
plot(xdata,time_fast,'rd')
P = polyfit(xdata,time_fast,1);
yfit = P(1)*xdata+P(2);
hold on;
plot(xdata,yfit,'k-');
legend('data','linear fit')