function x=qs_shifted_function_blocks(p,q,a,g,h,b,d,sigma,y)

% This is companion software to the manuscript
% "Efficient solution of parameter dependent
% quasiseparable systems",
% by P. Boito, Y. Eidelman and L. Gemignani.
%
% This code computes the solution of shifted linear systems
% of the form (A+sigma_i I)x_i = y, i=1,...,M,
% where A is quasiseparable.
%
% Here A can be a block quasiseparable matrix,
% that is, the m_i's need not be 1. We assume
% for simplicity that they are all equal to bsize.
%
% Input:
% - lower generators p,q,a
% - upper generators g,h,b
% - diagonal entries d
% - shifts sigma.
% - block vector (matrix) y.
% Output:
% - solution vectors x.
%
% Please note: no check is made on the consistency of the sizes
% of the input arrays.
%
% Last modification on October 20, 2016.

% Set up initial data. We use the following sizes:
% p is (bsize,rL,N)  but first one not used
% q is (rL,bsize,N-1)
% a is (rL,rL,N-1)  but first matrix not used
% g is (bsize,rU,N-1)
% h is (rU,bsize,N)   but first one not used
% b is (rU,rU,N-1)    but first matrix not used
% d is (bsize,bsize,N)
% y is (bsize,N)    storing blocks horizontally

[bsize,N]=size(y); % bsize = block size, N = size of system
M=length(sigma);  % number of shifts
[~,rL,~]=size(p); % lower qs order
[rU,~,~]=size(h); % upper qs order

m=bsize*ones(N,1);
w=zeros(bsize,N);

% Initialize matrix of solutions.
x=zeros(N*bsize,M);

% Implement algorithm.

% Start with the factorization A + sigma I = V * T:
% initialize generators for V
pV=zeros(bsize,rL,N);
qV=zeros(rL,bsize,N-1);
aV=zeros(rL,rL,N-1);
gV=zeros(bsize,rU,N-1);
hV=zeros(rU,bsize,N);
bV=zeros(rU,rU,N-1);
dV=zeros(bsize,bsize,N);

% 1 %

% Compute sizes:
rho=zeros(N,1);
rhop=zeros(N,1);
nu=zeros(N,1);
for k=N:-1:2
    rho(k-1)=min(m(k)+rho(k),rL);
end
for k=1:N-1
    rhop(k)=rho(k)+rU;
end
nu(1)=m(1)+rho(1);
for k=2:N
    nu(k)=m(k)+rho(k)-rho(k-1);
end

gT=zeros(bsize,rhop(1),N);
hT=zeros(rhop(1),bsize,N);
bT=zeros(rhop(1),rhop(1),N-1);
dT=zeros(bsize,bsize,N);

[V,R]=qr(p(:,:,N));
V_N=V;
Xold=R(1:rho(N-1),1:rL);

pV(1:m(N),1:rho(N-1),N)=V_N(1:m(N),1:rho(N-1));
if m(N)>rho(N-1)
    dV(1:m(N),1:nu(N),N)=V_N(1:m(N),rho(N-1)+1:end); 
    % otherwise it is empty
end

auxmat=[h(:,:,N);V_N'*d(:,:,N)];
hT(1:rhop(N-1),1:m(N),N)=auxmat(1:rhop(N-1),:);
if m(N)+rU>rhop(N-1)
    dT(1:nu(N),1:m(N),N)=auxmat(rhop(N-1)+1:rhop(N-1)+nu(N),:);
    % else dT(N) is empty
end

auxc=V_N'*y(:,N);
if m(N)>rho(N-1) 
    c=auxc(1:rho(N-1));
    w(1:nu(N),N)=auxc(rho(N-1)+1:end);
else
% w(N) is empty
    c=auxc;
end

gamma_N=[zeros(rU,m(N));pV(1:m(N),1:rho(N-1),N)'];
gamma=zeros(2*rU,bsize,N-1);

for k=N-1:-1:2
    % 1.2.1
    emme=[p(:,:,k);Xold*a(:,:,k)];
    [V,R]=qr(emme);
    
    Xnew=R(1:rho(k-1),:);
    pV(1:m(k),1:rho(k-1),k)=V(1:m(k),1:rho(k-1));
    aV(1:rho(k),1:rho(k-1),k)=V(m(k)+1:m(k)+1+rho(k)-1,1:rho(k-1));
    if rho(k)+m(k)>rho(k-1)
        dV(1:m(k),1:nu(k),k)=V(1:m(k),rho(k-1)+1:rho(k-1)+nu(k));
        qV(1:rho(k),1:nu(k),k)=V(m(k)+1:m(k)+rho(k),rho(k-1)+1:rho(k-1)+nu(k));
    end
    
    %1.2.2
    W=[h(:,:,k),b(:,:,k),zeros(rU,rho(k));d(:,:,k),g(:,:,k),zeros(m(k),rho(k));...
        Xold*q(:,:,k),zeros(rho(k),rU),eye(rho(k))];
    Z=[eye(rU),zeros(rU,rho(k)+m(k));zeros(rho(k)+m(k),rU),V']*W;
    hT(1:rhop(k-1),1:m(k),k)=Z(1:rhop(k-1),1:m(k));
    bT(1:rhop(k-1),1:rhop(k),k)=Z(1:rhop(k-1),m(k)+1:rhop(k)+m(k));
    if (rU+rho(k)+m(k))>rhop(k-1)
        dT(1:nu(k),1:m(k),k)=Z(rhop(k-1)+1:rhop(k-1)+nu(k),1:m(k));
        gT(1:nu(k),1:rhop(k),k)=Z(rhop(k-1)+1:rhop(k-1)+nu(k),m(k)+1:rhop(k)+m(k));
    end
    
    gamma(1:rU+rho(k-1),1:m(k),k)=[zeros(rU,m(k));pV(1:m(k),1:rho(k-1),k)'];
    auxv=V'*[y(:,k);c];
    c=auxv(1:rho(k-1));
    w(1:nu(k),k)=auxv(rho(k-1)+1:rho(k-1)+nu(k));
    
    Xold=Xnew;
end
% 1.3 (larger sizes here)
V1=eye(nu(1));
dT1=[d(:,:,1);Xold*q(:,:,1)];    % CHANGED THIS!!!!!!
gT1=[g(:,:,1) zeros(m(1),rho(1)); zeros(rho(1),rU) eye(rho(1))];
gamma1=[eye(m(1));zeros(rho(1),m(1))];
w1=V1'*[y(:,1);c];

% 2 %

for i=1:M   % for each shifted system
    
    dR=zeros(bsize,bsize,N);
    gR=zeros(bsize,max(rhop),N-1);
    v=zeros(bsize,N);
    
    % 2.1.1
    [U,R]=qr(dT1+sigma(i)*gamma1);
    dR(1:m(1),1:m(1),1)=R(1:m(1),1:m(1));
    
    auxv=U'*gT1;
    auxw=U'*w1;
    gR(1:m(1),1:rhop(1),1)=auxv(1:m(1),:);
    Y=auxv(m(1)+1:end,:);
    v(:,1)=auxw(1:m(1));
    alpha=auxw(m(1)+1:end);
    
    % 2.1.2
    
    for k=2:N-1
        if (rU+rho(k)+m(k))>rhop(k-1)  % that is, if dT(k) and dV(k) are not empty
            emme=[Y*(hT(:,:,k)+sigma(i)*gamma(:,:,k));dT(:,:,k)+sigma(i)*dV(:,:,k).'];
        else
            emme=Y*(hT(1:rhop(k-1),:,k)+sigma(i)*gamma(1:rhop(k-1),:,k));
        end
        [U,R]=qr(emme);
        dR(:,:,k)=R(1:m(1),1:m(1));
        
        if (rU+rho(k)+m(k))>rhop(k-1) % that is, if gT is not empty
            auxv=U'*[Y*bT(:,:,k); gT(:,:,k)];
            auxw=U'*[alpha; w(:,k)];
        else
            auxv=U'*Y*bT(1:rhop(k-1),1:rhop(k),k);
            auxw=U'*alpha;
        end
        
        gR(1:m(k),1:rhop(k),k)=auxv(1:m(k),1:rhop(k));
        v(:,k)=auxw(1:m(k));
        Y=auxv(m(k)+1:m(k)+rho(k),1:rhop(k));
        alpha=auxw(m(k)+1:m(k)+rho(k));
    end
    
    %2.1.3
    if m(N)+rU>rhop(N-1)
        auxmat=[Y*(hT(1:rhop(N-1),:,N)+sigma(i)*gamma_N);...
            dT(1:nu(N),1:m(N),N)+sigma(i)*dV(1:m(N),1:nu(N),N)'];
    else
        % dT is empty
        auxmat=Y*(hT(1:rhop(N-1),:,N)+sigma(i)*gamma_N);
    end
    [U,R]=qr(auxmat);
    dR(:,:,N)=R(1:m(N),1:m(N));
    if m(N)>rho(N-1)
       v(:,N)=U'*[alpha;w(1:nu(N),N)];
    else
       v(:,N)=U'*alpha;
    end
    
    %2.2 (compute solution x_i=x(:,i))
    
    %2.2.1
    eta=zeros(rhop(1),N-1);
    x((N-1)*bsize+1:N*bsize,i)=dR(:,:,N)\v(:,N);
    eta(1:rhop(N-1),N-1)=(hT(1:rhop(N-1),:,N)+sigma(i)*gamma_N)*x((N-1)*bsize+1:N*bsize,i);
    %2.2.2
    
    for k=N-1:-1:2
        x((k-1)*bsize+1:k*bsize,i)=dR(:,:,k)\(v(:,k)-gR(:,1:rhop(k),k)*eta(1:rhop(k),k));
        eta(1:rhop(k-1),k-1)=bT(1:rhop(k-1),1:rhop(k),k)*eta(1:rhop(k),k)+...
            (hT(1:rhop(k-1),:,k)+sigma(i)*gamma(1:rhop(k-1),:,k))*x((k-1)*bsize+1:k*bsize,i);
    end
    %2.2.3
    x(1:bsize,i)=dR(:,:,1)\(v(:,1)-gR(:,:,1)*eta(:,1));
end

end